//  CIM - Click-It Master
//  Preferable for mono audio tracks but will work with stereo
//

string name="Click-It Master";
string description="Programmable Pattern Metronome Plug-in";
string author="David Baer";

// Input Parameters:
//
//   0 - Dummy1
//       0 to 100%
//   0 - Dummy2
//       0 to 100%

array<string> inputStringsNames={"Cntl File Name"};
array<string> inputStrings(1);

array<string> inputParametersNames={"Control"};
array<string> inputParametersUnits={""};
array<double> inputParameters(inputParametersNames.length);
array<double> inputParametersMin={0};
array<double> inputParametersMax={3};
array<double> inputParametersDefault={1};
array<int> inputParametersSteps={4};
array<string> inputParametersEnums={"Reset;Off;Record;Record/Play"};

int debugLvl = 0;
double debugCallCntStart = 0;
double debugCallCntStop = 99999999999999;
double callCnt = 0;

int onLvl = 0; // 0-off, 1-rec, 2-rec+play
int lastControlValue = -1;
const double invSampleRate = 1.0 / sampleRate;
double lastBpm = 0;
double bps;
double inv60 = 1.0 / 60.0;
int lastTimeSigTop = 0;
int lastTimeSigBottom = 0;
int currTimeSigTop;
int currTimeSigBottom;
int currTType;
int currTOffset;
int currTPlaybackPos = 9999999;
int currTSampleCnt;
int timeSigComposit;
int tickNumber = -1;
double qnPerSample;
double qnPerMeasure;
int ticksPerMeasure;
int ticksPerQn;
int64 lastPositionInSamples;
int lastSamplesToProcess;
double qnPerTick;
double nextTick;    // units of QN
double lastTick;    // units of QN

string nonOperationReason = "";
bool outOfService = false;

string homeDir = "";
string filePathSlashChar = "/";
string audioDir = "";
string cntlFileName = "";
bool cntlFileInitialReadDone = false;
string beatMap;
int timeSigBeats;
int timeSigBeatUnits;

int tsCnt = 0;
array<int> tsKey(64);
array<int> tsTickCnt(64); // i.e. ticks per measure
array<int> tsTOffset(64);
int tCnt = 0;
array<int> t(1028);

array<int> tSampleCnt(3);
array<array<double>> tSamples(3, array<double>(20000));

bool initialize()
{
	string s = userDocumentsPath;
   if (s.findFirst("/") == -1)
      filePathSlashChar = "\\";
   homeDir = s + "SoundBytes Magazine" + filePathSlashChar +
   		"Click-It Master" + filePathSlashChar;
   audioDir = homeDir + "44100" + filePathSlashChar;
   return true;
}

void processBlock(BlockData& data)
{
   int tickOffset;
   double currBpm;
   int samplesToProcess;
   int64 positionInSamples;
   bool resuming;
   double positionInQuarterNotes;
   double currentMeasureDownBeat;
   double qnInCurrBuffer;
   double deltaQn;
   
   if(@data.transport is null)
   {
   	nonOperationReason = "data.transport is null";
   	outOfService = true;
   	return;	
   }

   if (outOfService)
   {
      if (nonOperationReason != "")
      {
         print(nonOperationReason);
         nonOperationReason = "";
      }
      return;
   }
   bool proceed = ((data.transport.isRecording && (onLvl > 0)) ||
   		(data.transport.isPlaying && (onLvl > 1)));
   if (!proceed)
   {
      tickNumber = -1;
      return;
   }
   positionInSamples = data.transport.positionInSamples;
   samplesToProcess = data.samplesToProcess;

   callCnt++;
   if (debugLvl > 1)
   {
      print(">>>>" + callCnt + 
            "  samplesToProcess=" + data.samplesToProcess + "  " +
            data.transport.timeSigTop + "/" +
            data.transport.timeSigBottom + 
            "  bpm=" + data.transport.bpm);
      print("    positionInQuarterNotes=" +
            data.transport.positionInQuarterNotes +
            "  currentMeasureDownBeat=" +
            data.transport.currentMeasureDownBeat);
      print("    positionInSamples=" + positionInSamples);
   }
   
   currBpm = data.transport.bpm;
   currTimeSigTop = data.transport.timeSigTop;
   currTimeSigBottom = data.transport.timeSigBottom;
   positionInQuarterNotes = data.transport.positionInQuarterNotes;
   currentMeasureDownBeat = data.transport.currentMeasureDownBeat;
   resuming = (((lastPositionInSamples + lastSamplesToProcess) !=
   		positionInSamples) || (tickNumber == -1));
   lastSamplesToProcess = samplesToProcess;
   lastPositionInSamples = positionInSamples;
   
   if (resuming)
   {
      calcTimeSigAttributes();
      deltaQn = positionInQuarterNotes - currentMeasureDownBeat;
      tickNumber = int(deltaQn * ticksPerQn);
      lastTick = currentMeasureDownBeat + (tickNumber* qnPerTick);
      nextTick = lastTick + qnPerTick;
      currTSampleCnt = 0;
      currTPlaybackPos = 9999999;
      for (int i = 0; i < samplesToProcess; i++)
         data.samples[0][i] = 0;
      if (debugLvl > 0)
      {
         print("**************************************************");
         print("Starting/resuming");
         print("positionInQuarterNotes=" + positionInQuarterNotes +
         	   "  currentMeasureDownBeat=" + currentMeasureDownBeat);
         print("deltaQn=" + deltaQn + "  tickNumber=" + tickNumber +
         	   "  lastTick=" + lastTick +
         	   "  nextTick=" + nextTick);
      }
   }

   if (currTPlaybackPos < currTSampleCnt)
      fillSamples(data.samples, samplesToProcess, 0);
   
   if (lastBpm != currBpm)
   {
      lastBpm = currBpm;
      bps = data.transport.bpm * inv60;
      qnPerSample = bps * invSampleRate;
      if (debugLvl > 0)
      {
         print("**************************************************");
         print("Bpm Change to " + currBpm);
      }
   }

   qnInCurrBuffer = samplesToProcess * qnPerSample;
   
   if (nextTick <= positionInQuarterNotes + qnInCurrBuffer)
   {
      deltaQn = nextTick - positionInQuarterNotes;
      double deltaTicks = (deltaQn / bps) * sampleRate;
      tickOffset = int(deltaTicks);
      if (debugLvl > 0)
      {
         print("**************************************************");
         print("Tick-in-Buffer  lastTick=" + lastTick + 
         		"  nextTick=" + nextTick +
               "  qnInCurrBuffer=" + qnInCurrBuffer);
         print("positionInQuarterNotes=" + positionInQuarterNotes +
         	   "  currentMeasureDownBeat=" + currentMeasureDownBeat +
         	   "  ticksPerQn=" + ticksPerQn);
         print("deltaQn=" + deltaQn + 
         	   "  deltaTicks=" + deltaTicks + "  tickOffset=" + tickOffset);
         double tickTime = data.transport.positionInSeconds +
         	   (tickOffset / sampleRate);
         print("positionInSeconds=" + data.transport.positionInSeconds +
               "  tickOffset=" + tickOffset +
         	   "  tickTime=" + tickTime);
      }
      lastTick = nextTick;
      tickNumber++;
      if (tickNumber == ticksPerMeasure)
      	tickNumber = 0;
      nextTick = lastTick + qnPerTick;
      deltaQn = nextTick - int(nextTick);
      if (abs(deltaQn) < 0.0001)
      	nextTick = floor(nextTick + 0.01);
      if (tickNumber == ticksPerMeasure) // new measure
      	tickNumber = 0;
      setUpTickPlayback();
      fillSamples(data.samples, samplesToProcess, tickOffset);
      if (debugLvl > 0)
      	print("tickNumber=" + tickNumber + "  lastTick=" + lastTick +
      			"  nextTick=" + nextTick);				
   }
   if ((currTimeSigTop != lastTimeSigTop) ||
         (currTimeSigBottom != lastTimeSigBottom))
      calcTimeSigAttributes();
}

void updateInputParametersForBlock(const TransportInfo@ info)
{
   int control = int(inputParameters[0] + 0.5);
   if ((control != lastControlValue) ||
   		(!cntlFileInitialReadDone))
   {
      onLvl = control - 1;
      if (((control == 0) && (lastControlValue != 0)) ||
      	(!cntlFileInitialReadDone))
      {
      	cntlFileInitialReadDone = true;
         resetOperation();
      }
   }
   lastControlValue = control;
}

void calcTimeSigAttributes()
{
   int i;
   
   timeSigComposit = (currTimeSigTop * 256) + currTimeSigBottom;
   lastTimeSigTop = currTimeSigTop;
   lastTimeSigBottom = currTimeSigBottom;
   bool found = false;
   for (i = 0; i < tsCnt; i++)
   {
      if (tsKey[i] == timeSigComposit)
      {
         found = true;
         break;
      }
   }
   if (found)
   {
      ticksPerMeasure = tsTickCnt[i];
      currTOffset = tsTOffset[i];
   }
   else
   {
      ticksPerMeasure = currTimeSigTop;
      currTOffset = 0;
   }
 
   qnPerMeasure = (currTimeSigTop * 4.0) / currTimeSigBottom;
   ticksPerQn = int((ticksPerMeasure / qnPerMeasure) + 0.001);
   qnPerTick = 1.0 / ticksPerQn;
   qnPerSample = bps * invSampleRate;
   nextTick = lastTick + qnPerTick;
   if (debugLvl > 0)
   {
      print("**************************************************");
      print("New-Time-Sig: " +
         currTimeSigTop + "/" + currTimeSigBottom +
         "  qnPerMeasure=" + qnPerMeasure +
         "  ticksPerMeasure=" + ticksPerMeasure);
      print("lastTick=" + lastTick + "  nextTick=" + nextTick);
      string s = "";
      for (i = 0; i < ticksPerMeasure; i++)
         s = s + t[currTOffset + i];
      print("timeSigComposit=" + timeSigComposit + "  Tick-Map=|" + s + "|");
   }
}

void setUpTickPlayback()
{
   currTType = t[currTOffset + tickNumber];
   if (currTType > -1)
   {
      currTSampleCnt = tSampleCnt[currTType];
      currTPlaybackPos = 0;
   }
   else
   {
      currTSampleCnt = 0;
      currTPlaybackPos = 9999999;
   }	
}

void resetOperation()
{
   outOfService = false;
   nonOperationReason = "";
   readControlFile();
   loadTickSound(0, audioDir + "DownbeatClick.wav");
   loadTickSound(1, audioDir + "PrimaryClick.wav");
   loadTickSound(2, audioDir + "SecondaryClick.wav");
}

void loadTickSound(const int tickType, const string &in fileName)
{
   WaveFileReader wfr;
   array<double> sample(1);  
   int i;
    
   tSampleCnt[tickType] = 0;
   if (wfr.openFile(fileName, 1))
   {
      for (i = 0; i < 20000; i++)
      {
         if (wfr.isDone())
            break;
         wfr.readSample(sample);
         tSamples[tickType][i] = sample[0];
      }
      tSampleCnt[tickType] = i;
      wfr.close();
   }
   else
      print("Error - tick audio file open problem: " + fileName);
}

void readControlFile()
{
   file f;
   
   tsCnt = 0; // first set up default tick pattern
   t[0] = 0; // downbeat tick
   for (int i = 1; i < 64; i++)
      t[i] = 2; // secondary beat tick
   tCnt = 64;

   string s = inputStrings[0] + " ";
   while ((s.length > 0) && (s[0] == ' '))
   {
      s = s.substr(1);
   }
   if (s.length == 0)
   {
      s = "CIM.cntl";
   }
   else
   {
      s = s.substr(0, s.findFirst(" "));
   }
   cntlFileName = homeDir + s;
   if (f.open(cntlFileName, "r") >= 0)
   {
      while (!f.isEndOfFile())
      {
         s = f.readLine();
         int i = s.length;
         if (s.substr(i - 2) == "\r\n")
         {
            s = s.substr(0, i - 2);
         }
         else if (s.substr(i - 1) == "\n")
         {
            s = s.substr(0, i - 1);
         }
         i = s.findFirst("\t");
         while (i >= 0)
         {
            s[i] = ' ';
            i = s.findFirst("\t");
         }
         if (debugLvl > 0)
            print("Cntl-Rcd=|" + s + "|");
         int j = s.length;
         for (i = 0; i < j; i++)
         {
            if (s[i] != ' ')
               break;
         }
         if (i == j)
            break;
         if (s.substr(0, 6) == "Debug ")
         {
            parseDebugLvlDefinition(s);
         }
         else if (s.substr(0, 6) == "Audio ")
         {
            parseAudioDirDefinition(s);
         }
         else
         {
            if (!parseTimeSigDefinition(s))
               print("Invalid cntl record: " + s);
         }
      }
      f.close();
   }
   else
   {
      nonOperationReason = "Unable to open " + cntlFileName;
      outOfService = true;
      return;
   }
   
   if (debugLvl > 0)
   {
      print("debugLvl=" + debugLvl +
            "  debugCallCntStart=" + debugCallCntStart +
            "  debugCallCntStop=" + debugCallCntStop);   
      print("audioDir=" + audioDir);
   	print("tsCnt=" + tsCnt);
      for (int i = 0; i < tsCnt; i++)
      {
         s = "|" + tsKey[i] + "|" + tsTickCnt[i] + "|" +
               tsTOffset[i] + "|";
         int j = tsTOffset[i];
         int k = tsTickCnt[i];
         for (int l = 0; l < k; l++)
            s = s + t[j + l];
         s = s + "|";
	      print(s);
      }
   }
}

void parseDebugLvlDefinition(const string &in def)
{
   uint ui;
   int i;
   string s = def.substr(6);
   i = int(parseInt(s, 10, ui));
   if ((ui < 1) || (i < 0) || (i > 2))
      return;
   debugLvl = i;
   s = s.substr(ui + 1);
   i = int(parseInt(s, 10, ui));
   if ((ui < 1) || (i < 0))
      return;
   debugCallCntStart = i;
   s = s.substr(ui + 1);
   i = int(parseInt(s, 10, ui));
   if ((ui < 1) || (i < 0))
      return;
   debugCallCntStop = i;
}

void parseAudioDirDefinition(const string &in def)
{
   int i;
   audioDir = homeDir + "44100" + filePathSlashChar;
   string s = def.substr(6) + " ";
   while ((s.length > 0) && (s[0] == ' '))
   {
      s = s.substr(1);
   }
   if (s.length == 0)
      return;
   audioDir = homeDir + s.substr(0, s.findFirst(" ")) +
   		filePathSlashChar;
}


bool parseTimeSigDefinition(const string &in def)
{
   string s = "";
   int i = 0;
   uint ui = 0;
   
   if (debugLvl > 0)
      print("TS-Def=|" + def + "|");
   s = def + "                                        " +
         "                                        ";
   timeSigBeats = int(parseInt(s, 10, ui));
   if (ui < 1)
      return false;
   s = s.substr(ui + 1);
   timeSigBeatUnits = int(parseInt(s, 10, ui));
   if (ui < 1)
      return false;
   s = s.substr(ui + 1);
   ticksPerMeasure = int(parseInt(s, 10, ui));
   if (ui < 1)
      return false;
   s = s.substr(ui + 1);
   i = s.findFirst("!");

   if (i < 0)
      return false;
   beatMap = s.substr(i, ticksPerMeasure);
   if ((timeSigBeats < 1) || (timeSigBeats > 64))
      return false;
   if ((timeSigBeatUnits != 1) &&
         (timeSigBeatUnits != 2) &&
         (timeSigBeatUnits != 4) &&
         (timeSigBeatUnits != 8) &&
         (timeSigBeatUnits != 16) &&
         (timeSigBeatUnits != 32))
      return false;
   if (ticksPerMeasure > 64)
      return false;
   if (tsCnt == 64)
   {
      print("Too many time signitures: " + def);
      return false;
   }
   if ((tCnt + ticksPerMeasure) > 1024)
   {
      print("Too many beat definitions: " + def);
      return false;
   }
   loadTickData();
   return true;
}

void loadTickData()
{
   timeSigComposit = (timeSigBeats * 256) + timeSigBeatUnits;
   for (int i = 0; i < tsCnt; i++)
   {
      if (tsKey[i] == timeSigComposit)
			return;
	}
   tsKey[tsCnt] = timeSigComposit;
   tsTickCnt[tsCnt] = ticksPerMeasure;
   tsTOffset[tsCnt] = tCnt;
   for (int i = 0; i < ticksPerMeasure; i++)
   {
      string c = beatMap.substr(i, 1);
      if (c == "!") 
         t[tCnt] = 0;
      else if (c == "+")
         t[tCnt] = 1;
      else if (c == "-")
         t[tCnt] = 2;
      else
         t[tCnt] = -1;
      tCnt++;
   }
   tsCnt++;
}

void fillSamples(array<array<double>> &samples, const int samplesToProcess,
		const int startPos)
{
   if (debugLvl > 0)
   {
      print(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
      print("FillSamples  samplesToProcess=" + samplesToProcess +
            "  startPos=" + startPos);
      print("currTType=" + currTType +
            "  currTPlaybackPos=" + currTPlaybackPos + "<<<<<<<<<<<<<");
   }
   if (currTType == -1)
   {
      currTPlaybackPos = 9999999;
   	return;
   }
   for (int i = startPos; i < samplesToProcess; i++)
   {
      samples[0][i] = tSamples[currTType][currTPlaybackPos];
      currTPlaybackPos++;
      if (currTPlaybackPos == currTSampleCnt)
      {
         currTPlaybackPos = 9999999;
         currTType = -1;
         break;
      }
   }
   // propogate to additional channels if not mono
   for (uint i = 1; i < audioOutputsCount; i++)
   {
      for (int j = 0; j < samplesToProcess; j++)
      {
         samples[i][j] = samples[0][j];
      }
   }
}

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
// The following class was lifted from Wavefile.hxx and modified
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

class WaveFileReader
{
    /** open a file for streaming audio data from it.
    *   maxChannelsCount is the maximum number of channels that will be read from the
    *   file. Set to -1 if you intend to use the number of channels of the file.
    */
    bool openFile(string filePath, int maxChannelsCount=-1)
    {
        bool ok=false;
        if(f.open(filePath,"r")>=0)
        {
            if(header.read(f))
            {
                ok=true;
                blockSize=header.channelsCount*header.bytesPerSample;
					 samplesCount = header.samplesCount;
                if(header.bytesPerSample>1 && header.bytesPerSample<4)
                {
                    uint64 maxValue=(1<<(header.bytesPerSample*8-1))-1;
                    gainFactor=0.99/double(maxValue);
                }
                if(maxChannelsCount>0 && uint(maxChannelsCount)<header.channelsCount)
                {
                    channelsToRead=maxChannelsCount;
                    channelsToSkip=(header.channelsCount-maxChannelsCount);
                }
                else
                {
                    channelsToRead=header.channelsCount;
                    channelsToSkip=0;
                }
            }
        }
        samplesRead = 0;
        return ok;
    }

    /** Read one sample from the current position in the file.
    *   Assumes that the file was succesfuly opened.
    *   oSample is an array with at least maxChannelsCount (@see openFile),
    *   or the number of channels containedin the file (@see get_channelsCount).
    */
    void readSample(array<double>& oSample)
    {
        switch(header.bytesPerSample)
        {
            // 8-bit wav file contain only positive values.
        case 1:
            {
                for(uint ch=0;ch<channelsToRead;ch++)
                {
                    oSample[ch]=(double(f.readUInt(header.bytesPerSample))-128.0)/128.0;
                }
                for(uint ch=0;ch<channelsToSkip;ch++)
                {
                    f.readUInt(header.bytesPerSample);
                }

                break;
            }
            // single precision floating point wav file
        case 4:
            {
                for(uint ch=0;ch<channelsToRead;ch++)
                {
                    oSample[ch]=f.readFloat();
                }
                for(uint ch=0;ch<channelsToSkip;ch++)
                {
                    f.readFloat();
                }
                break;
            }
        case 8:
            // double precision floating point wav file
            {
                for(uint ch=0;ch<channelsToRead;ch++)
                {
                    oSample[ch]=f.readDouble();
                }
                for(uint ch=0;ch<channelsToSkip;ch++)
                {
                    f.readDouble();
                }
                break;
            }
            // 16 or 24-bit integer wav file
        default:
            {
                for(uint ch=0;ch<channelsToRead;ch++)
                {
                	  int i = f.readInt(header.bytesPerSample);
                    oSample[ch]=double(i)*gainFactor;
                }
                for(uint ch=0;ch<channelsToSkip;ch++)
                 {
                     f.readInt(header.bytesPerSample);
                 }
                 break;
            }
        }
        samplesRead++;
    }
    /** Move file cursor to sample number "samplePosition".
    *   Warning: does not check if end of data chunk is reached.
    */
    bool setPos(uint samplePosition)
    {
        return f.setPos(header.headerSize+samplePosition*blockSize)>=0;
    }

    /** Move file cursor by "sampleOffset" samples.
    *   returns false if failed.
    *   Warning: does not check if end of data chunk is reached.
    */
    bool movePos(int sampleOffset)
    {
        int newPos=f.getPos()+sampleOffset*blockSize;
        if(newPos>=int(header.headerSize))
            return f.setPos(sampleOffset*blockSize)>=0;
        else
            return false;
    }

    /** Check if end of file reached.
    *
    */
    bool isDone()
    {
        return (f.isEndOfFile() || (samplesRead >= samplesCount));
    }

    /** close the audio file.
    *
    */
    bool close()
    {
        return f.close()>=0;
    }

    /** channelsCount property (read only).
    *
    */
    int get_channelsCount()const
    {
        return header.channelsCount;
    }

    // private data
    private file    f; // the file
    private uint    blockSize=0; // sample block size (all channels)
    private uint    channelsToRead=0; // number of channels to read from file
    private uint    channelsToSkip=0; // number of channels to skip (channels not used)
    private double gainFactor=1; // gain factor for integer files
    private WaveFileHeader header; // the wave file header data
    private uint64 samplesCount = 0;
    private uint64 samplesRead = 0;
};

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
// The following class was lifted from Wavefile.hxx and modified
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

class WaveFileHeader
{
    uint64  channelsCount=0;        // number of audio channels
    uint64  samplesCount=0;         // number of audio samples
    uint64  sampleRate=0;           // the sample rate of the audio data
    uint    bytesPerSample=0;       // the number of bytes per individual channel sample      
    uint    headerSize=0;           // the size of the header (offset to audio data)

    /** Read header data from file.
    *   Assumes that the file is at position 0.
    */
    bool read(file& f)
    {
        bool ok=false;
        string tempString;
        tempString=f.readString(4);
        if(tempString=="RIFF")
        {
            int64 size=f.readUInt(4);
            tempString=f.readString(4);
            if(tempString=="WAVE")
            {
                // looking for fmt subchunk => skipping all others 
                tempString=f.readString(4); // data id
                while(tempString!="fmt " && !f.isEndOfFile())
                {
                    uint64 dataSize = f.readUInt(4);
                    f.movePos(dataSize);        // skip junk data
                    tempString=f.readString(4); // read next chunk id
                }
                // should now be at fmt chunk
                uint64 fmtSize = f.readUInt(4);
                if(fmtSize>=16) // PCM format size is 16 or 18
                {
                    uint64 dataOffset=fmtSize-16;
                    uint64 format=f.readInt(2);
                    if(format==1 || format==3) //uncompressed PCM format or float are supported
                    {
                        channelsCount=f.readUInt(2);
                        sampleRate=f.readUInt(4);
                        uint64 byteRate=f.readUInt(4);
                        uint64 blockSize=f.readUInt(2);
                        bytesPerSample = f.readUInt(2)/8;
                        // end of subchunk
                        f.movePos(dataOffset);

                        // looking for data subchunk => skipping all others 
                        tempString=f.readString(4); // data id
                        while(tempString!="data" && !f.isEndOfFile())
                        {
                            uint64 dataSize = f.readUInt(4);
                            f.movePos(dataSize);        // skip junk data
                            tempString=f.readString(4); // read next chunk id
                        }
                        if(tempString=="data")
                        {
                            uint64 dataSize = f.readUInt(4);
                            samplesCount=dataSize/blockSize;
                        }
                        headerSize=f.getPos();
                        ok=true;
                    }
                }
            }
        }
        return ok;
    }

};

